# -*- cperl -*-

#  This file is part of PuDaC.

#  PuDaC is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.

#  PuDaC is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with PuDaC  If not, see <http://www.gnu.org/licenses/>.

# This file contains the %objectstranslations hash of object-to-C translations

## Notes:
## event objects-
##  data flows down (by calls)
##  each call contains payload
##  certain objects source events (metro, delay, midiin, netreceive, &c)
## DSP objects-
##  data flows up (by return)
##  objects cache with the sample # include
##  certain objects sink data (dac~, tabsend~, tabwrite~, print~, &c)

## All objects: inletnumber >=0 -- message intlets (number N = (N+1)th inlet)
##  inletnumber < 0 -- dsp outlet (number -N = Nth outlet)

my $MSGwarning = 0;
my $ROUTEwarning = 0;

#use enum qw(:PDT_ DSP FLOAT SYMBOL BANG NONDSP);


%objecttranslations =
    ('b' => { 'SYNONYM' => 'bang' },
     'f' => { 'SYNONYM' => 'float' },
     'i' => { 'SYNONYM' => 'int' },
#    'l' => { 'SYNONYM' => 'lister' }, # from zexy, not implementing, but here for completeness.
     'r' => { 'SYNONYM' => 'receive' },
     'r~' =>{ 'SYNONYM' => 'receive~' },
     's' => { 'SYNONYM' => 'send' },
     's~' =>{ 'SYNONYM' => 'send~' },
     't' => { 'SYNONYM' => 'trigger' },
     'v' => { 'SYNONYM' => 'value' },
     'sel'=>{ 'SYNONYM' => 'select' },
     'fswap'=>{'SYNONYM'=> 'swap' },
     'hslider'=>{'SYNONYM'=>'hsl' },
     'radiobut'=>{'SYNONYM'=>'rdb' },
     'my_canvas'=>{'SYNONYM'=>'cnv' },
     'my_numbox'=>{'SYNONYM'=>'nbx' },
     'radiobutton'=>{'SYNONYM'=>'rdb' },

     # these are lies:
     'tabread4~' => { 'SYNONYM' => 'tabread~' },
     'tabread4' => { 'SYNONYM' => 'tabread' },
     'timer' => { 'SYNONYM' => 'realtime' },
     'cputime' => { 'SYNONYM' => 'realtime' },

     '+' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::+';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is an ADD object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) + takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft + lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft + lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) + doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) + takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) + only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     'print' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::print';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 my $initializer = "print";
	 if (length $arguments[0] > 0) {
	     $initializer = $arguments[0];
	 }

	 print qq[
/* object $number is a PRINT object */
void object$number (int inletnumber, const struct PDdata * message) {
    const char printname[] = {"$initializer"};
    const struct PDdata * args;

    if (inletnumber != 0) {
        error("(#$number) print only has one inlet");
        return;
    }

    args = message;
    printf("\%s: ",printname);
    while (args != NULL) {
        switch (args->type) {
        case PDC_TYPE_FLOAT:
            printf("\%f",args->val_float);
            break;
        case PDC_TYPE_BANG:
            printf("bang");
            break;
        case PDC_TYPE_SYMBOL:
            printf("\%s",args->val_symbol);
            /* Technically, we're supposed to clip at 78 characters. I think I can not care. */
            break;
        default:
            error("(#$number) unknown type \%d", args->type);
            return;
        }
        if (args->linked != NULL) {
            printf(" ");
        }
        args = args->linked;
        /* We assign this so that when our linked list runs out,
           the while up there is what terminates */
    }
    printf("\\n");
    return;
}
];
	 # print never has attached objects
     }
	      },
     'float' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::float';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 my $initializer = 0;
	 $initializer = $arguments[0]
	   if (exists $arguments[0]);

	 print qq[
/* object $number is a FLOAT object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float internal=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                internal = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) float for some reason cares that 2nd elt in list is nonfloat");
                        /* however this doesn't fail to propagate */
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = internal;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = internal;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) float doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) float takes only takes float in right inlet");
	        return;
            }
	    internal = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     'int'  => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::int';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 my $initializer = 0;
	 $initializer = $arguments[0]
	   if (exists $arguments[0]);

	 print qq[
/* object $number is an INT object */
void object$number (int inletnumber, const struct PDdata * message) {
    static int internal=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                internal = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) int for some reason cares that 2nd elt in list is nonfloat");
                        /* however this doesn't fail to propagate */
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = internal;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = internal;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) float doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) float takes only takes float in right inlet");
	        return;
            }
	    internal = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     'bang' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::bang';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 print qq[
/* object $number is a BANG object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the inlet */
            outvalue.type = PDC_TYPE_BANG;
            $attachedobjectcalls
            break;
	default: /* not an inlet */
	    error("(#$number) bang only has one inlet");
            return;
    }
    return;
}
];
}
	  },
     # metro loadbang, and so on are interesting, because they don't just involve
     # walking the trees.
     # loadbang is fairly straightforward, i think?
     # metronome, midiin, &c are much more interesting.
     'loadbang' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::loadbang';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 # basically we generate a bang object, but it only gets called by the main loop, once.
	 # i believe the order is ambiguous, PD 0.40 seems to do them in creation order.

	 print qq[
/* object $number is a LOADBANG object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue;

    /* loadbang should never be compiled in, except to the initial LB evaluation */

    outvalue.linked = NULL;

    outvalue.type = PDC_TYPE_BANG;
    $attachedobjectcalls
    return;
}
];

	 $loadbangs .= qq[
/* now call loadbang obj $number: */
object$number(PDC_CALLBACK,NULL); /* loadbang doesn't care about inlet # */
];
     }
		 },
     'metro' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::metro';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};
	 # here's where i'd like to have classes: i want both the object, and an accessor to
	 # its internal state so that i can find out if i should call it.

	 # options:
	 # 1) make inlet # -1 special
	 # 2) make the internal state not internal, but given its own unique global name

	 # i've done some of each here

	 my $initializer = 0;
	 $initializer = $arguments[0]
	   if (exists $arguments[0]);

#	 print STDERR "making metro $number w/ period $initializer ms\n";

	 print qq[
/* object $number is a METRO object */

struct timeval object${number}_time;
int object${number}_running;

void object$number (int inletnumber, const struct PDdata * message) {
    static float period_in_milliseconds=$initializer;
    struct PDdata outvalue;
    struct timeval TV_internal;

    outvalue.linked = NULL;
    outvalue.type = PDC_TYPE_BANG;

    switch (inletnumber) {
        case PDC_CALLBACK: /* a timeout */
            $attachedobjectcalls
            object${number}_time.tv_usec += period_in_milliseconds*1000;
            while (object${number}_time.tv_usec > 1000000 ) {
                object${number}_time.tv_sec++;
                object${number}_time.tv_usec-=1000000;
            }
            break;
	case 0: /* the toggly inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT: /* 0 means stop, non-zero means run */
                object${number}_running = (message->val_float != 0);
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) metro takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        /* schedule the next one */
                        period_in_milliseconds = message->linked->val_float;
                    }
                }
                break;
            case PDC_TYPE_BANG: /* bang means run */
                object${number}_running = 1;
                break;
            case PDC_TYPE_SYMBOL: /* "stop" means stop */
                if (strcasecmp("stop",message->val_symbol)==0) {
                    object${number}_running = 0;
                    return;
                }
                error("(#$number) metro doesn't understand that symbol (\%s)",message->val_symbol);
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    if (object${number}_running) {
                $attachedobjectcalls
                gettimeofday(&TV_internal,NULL);
                object${number}_time.tv_usec = TV_internal.tv_usec
                                               + period_in_milliseconds*1000;
                object${number}_time.tv_sec = TV_internal.tv_sec;
                while (object${number}_time.tv_usec > 1000000) {
                    object${number}_time.tv_sec++;
                    object${number}_time.tv_usec-=1000000;
                }
            }
	    break;
	case 1: /* the timer inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) metro takes only takes float in right inlet");
	        return;
            } else {
	        period_in_milliseconds = message->val_float;
                object${number}_running = 0;
            }
	    break;
	default: /* not an inlet */
	    error("(#$number) metro only has two inlets");
            return;
    }
    return;
}
/* }} cperl has trouble */
];

	 $globalinit .= qq[
  /* now initializing metro $number: */
  object${number}_running = 0;
];
	 $periteration .= qq[
/* Now checking timeout for obj $number: */
if (object${number}_running &&
    (now.tv_sec > object${number}_time.tv_sec ||
     (now.tv_sec == object${number}_time.tv_sec &&
      now.tv_usec > object${number}_time.tv_usec))) {
       object${number}(PDC_CALLBACK,NULL);
}
];
       }
		},
     'dac~' => { 'OT' => [], 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::dac~';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 croak "Non-stereo audio generation unimplemented!"
	   if ($#arguments > 0);

	 print qq[
/* object $number is a DAC~ object */
float *object${number}_upstr1, *object${number}_upstr2,
  *object${number}_cache1, *object${number}_cache2;

void object$number (int inletnumber, const struct PDdata * message) {
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL},
                         DSPtrace2={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache1 = NULL, * cache2 = NULL;
    float * writer;
    int i;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr1;
    if (NULL == DSPtrace2.dsp_stream) DSPtrace2.dsp_stream = object${number}_upstr2;
    if (NULL == cache1) cache1 = object${number}_cache1;
    if (NULL == cache2) cache2 = object${number}_cache2;

    writer = message->dsp_stream; /* break const-ness of message */

    if (inletnumber == PDC_DSP_DAC &&
        message->type == PDC_DSP_REQUEST) {
        if (message->sample_count != DSPtrace1.sample_count ||
            message->sample_count != DSPtrace2.sample_count) {
            memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
            memset(DSPtrace2.dsp_stream,0,block_size * sizeof(float));
            DSPtrace1.sample_count = DSPtrace2.sample_count = message->sample_count;
            $reverseDSPcalls
            memcpy(cache1,DSPtrace1.dsp_stream,sizeof(float)*block_size);
            memcpy(cache2,DSPtrace2.dsp_stream,sizeof(float)*block_size);
        }
        for (i=0;i<block_size;i++) {
            writer[i]+=cache1[i];
            writer[i+block_size]+=cache2[i]; /* This is ONLY OK in dac~s */
        }
    } else
        error("(#$number) dac~ got funny input? inlet:\%d type:\%d",inletnumber,message->type);
}
];#]

	 $globalinit .= qq[
  /* allocate "private" memory for dac~ $number: (dac~) */
  object${number}_upstr1 = malloc(block_size*sizeof(float));
  object${number}_upstr2 = malloc(block_size*sizeof(float));
  object${number}_cache1 = malloc(block_size*sizeof(float));
  object${number}_cache2 = malloc(block_size*sizeof(float));

  if (object${number}_upstr1 == NULL || object${number}_upstr2 == NULL ||
      object${number}_cache1 == NULL || object${number}_cache2 == NULL) {
      error("(#$number) Unable to allocate memory for obj $number (dac~)!");
      exit(1);
  }
];

	 $dsptrace .= qq[
    /* Checking on obj $number: (dac~) */
    object$number(PDC_DSP_DAC,&DSPtracer);
];
       }
	       },
     '\dspstub~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::\dspstub~';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 print qq[
/* object $number is a dsp stub, should never be rendered */
];

	 $globalinit .= qq[
  /* allocate "private" memory for obj $number */
];
	 $dsptrace .= qq[
  /* checking on dsp terminus obj $number */
];
       }
		       },
     'osc~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::osc~';
	 my ($number, %myself, @error) = @_;
	 my ($pitch) = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 if ($pitch != 0) {
	   $pitch = '= '.$pitch;
	 }

	 print qq[
/* object $number is an OSC~ object */
float *object${number}_upstr, *object${number}_cache;

void object$number (int inletnumber, const struct PDdata * message) {
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache = NULL;
    float * writer;
    static float cosph=0, freq $pitch;
    char * write_is_dsp;
    int i;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr;
    if (NULL == cache) cache = object${number}_cache;

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* find out if we have a dsp rate input */
            if (message->sample_count != DSPtrace1.sample_count) {
                memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
                DSPtrace1.sample_count = message->sample_count;
                $reverseDSPcalls
                if (DSPtrace1.is_dsp) {
                    for (i=0;i<block_size;i++) {
                        cache[i]=cos(cosph);
                        cosph+=DSPtrace1.dsp_stream[i] * M_PI*2.0/sample_rate;
                    }
                } else {
                    for (i=0;i<block_size;i++) {
                        cache[i]=cos(cosph);
                        cosph+=freq * M_PI*2.0/sample_rate;
                    }
                }
                while (cosph>7) cosph -= M_PI*2;
                while (cosph<-7) cosph += M_PI*2;
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 0: /* inlet 1 = freq */
            if (message->type == PDC_TYPE_FLOAT)
                freq = message->val_float;
            else
                error("(#$number) osc~ takes only floats or dsp chains in freq inlet (got \%d)",message->type);

            if (message->linked)
                if (message->linked->type == PDC_TYPE_FLOAT)
                    cosph = 0;
                else
                    error("(#$number) osc~ takes only float as 2nd elt of list in 1st inlet (got \%d)",message->type);

            break;
        case 1: /* inlet 2 = reset phase */
            /* only if numeric */
            if (message->type == PDC_TYPE_FLOAT)
                cosph = 0;
            else
                error("(#$number) osc~ takes only floats in reset_phase inlet (got \%d)",message->type);
            break;
        default:
            error("(#$number) osc~ only has two inlets (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for osc~ $number */
  object${number}_upstr = malloc(block_size*sizeof(float));
  object${number}_cache = malloc(block_size*sizeof(float));

  if (object${number}_upstr == NULL || object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (osc~)!");
    exit(1);
  }
];
       }
	       },
     'table' => { 'OT' => [], 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::table';
	 my ($number, %myself, @error) = @_;
	 my ($name,$size) = @{$myself{ARGS}};

	 # table without any arguments produces an incremententally named
	 # table (table0, table1, &c) of length 100

	 # table with one argument produces a table with name that argument,
	 # of length 100

	 $name = "table".($number_of_tables++) if (length $name < 1);
	 $size = 100 if ($size < 1);

	 # implemented:
         #  const resize print read

	 # Technically, tables are *intimately* tied into sends/receives.
	 # I'll care about these eventually:
	 #  CONST n (set to constant value)
	 #  RESIZE n (set array size)
	 #  PRINT (print size of table, only useful for debugging)
	 #  READ filename (doesn't resize) (whitespace-separated ascii floats)

	 #  SINESUM/COSINESUM size N1 N2 N3 .. Nn (Nth harmonic synthesis, also resizes)
	 #  NORMALIZE n (sets to maximum volume times n)
	 #  WRITE filename (ditto)
	 #  N1 N2 N3 .. Nn (set contents)

	 # The following don't make sense to implement:
	 #  BOUNDS
	 #  XTICKS
	 #  YTICKS

	 # The following is a pain to implement correctly so I'm not.
	 #  RENAME
	 #  (plus, i can't think of a legitimate use for it.)

	 print qq[
/* object $number is a TABLE object */
float *object${number}_table;
int object${number}_size;

void object$number (int inletnumber, const struct PDdata * message) {
  int i;
  float t;
  const struct PDdata * args;
  float ** tablecontents;
  int * tablelen;
  FILE * FH;

  if (inletnumber != PDC_SENT) {
    error("(#$number) table should have no connections (got event via inlet \%d)",inletnumber);
    return;
  }

  switch (message->type) {
    case PDC_TYPE_FLOAT:
      args = message;
      for (args=message,i=0;
           i<object${number}_size && args != NULL;
           i++,args=message->linked)
        object${number}_table[i]=message->val_float;
      return;
    case PDC_TYPE_SYMBOL:
      if (strcasecmp("const",message->val_symbol)==0) {
        if (message->linked == NULL) t=0;
        else if (message->linked->type != PDC_TYPE_FLOAT) {
          error("(#$number) table received 'const nonfloat', don't deal");
          return;
        } else t=message->linked->val_float;
        for (i=0;i<object${number}_size;i++)
          object${number}_table[i]=t;
        return;
      } else if (strcasecmp("resize",message->val_symbol)==0) {
        if (message->linked == NULL ||
            message->linked->type != PDC_TYPE_FLOAT) {
          error("(#$number) table received 'resize nonfloat', don't deal");
          return;
        }
        i = message->linked->val_float;
        if ( i < 1 ) i = 1;
        object${number}_table = realloc(object${number}_table,i);
        if (object${number}_table) {
          error("(#$number) TABLE REALLOC FAILED");
          exit(1);
        }
      } else if (strcasecmp("print",message->val_symbol)==0) {
        printf("table $name: \%d elts\\n",object${number}_size);
        return;
      } else if (strcasecmp("\\\\getguts",message->val_symbol)==0) {
        tablecontents = &(message->dsp_stream); /* needed for efficiency in all table routines */
        *tablecontents = object${number}_table;
        tablelen = &(message->sample_count);
        *tablelen = object${number}_size;
        return;
      } else if (strcasecmp("normalize",message->val_symbol)==0) {
      } else if (strcasecmp("read",message->val_symbol)==0) {
        if (message->linked && message->linked->type == PDC_TYPE_SYMBOL) {
          FH = fopen(message->linked->val_symbol,"r");
          if (FH == NULL) {
            error("(#$number) Table $name can't open file \%s\\n",message->linked->val_symbol);
            return;
          }

          for (i=0;
               i<object${number}_size &&
                   fscanf(FH,"\%g",&object${number}_table[i]) == 1;
               i++);
          if (i!=object${number}_size)
            error("(#$number) Only loaded \%d elts, not \%d.\\n",i,object${number}_size);
          fclose (FH);
        }
      } else if (strcasecmp("write",message->val_symbol)==0) {
      }
  }
}
];

	 $globalinit .= qq[
  /* Allocate shared memory for table $number */
  object${number}_size = $size;
  object${number}_table = calloc(object${number}_size,sizeof(float));
  if (object${number}_table == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (table)!");
    exit(1);
  }
  listens(\& object${number},"$name");
];
       }
		},
# to implement soundfiler, i must implement \msg
# to implement soundfiler, i need to deal with runtime names of objects
# so i may as well do send / receive next. sigh.
     'send' => { 'OT' => [], 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::send';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};

	 my $name = $arguments[0];

	 # two kinds of sends:
	 # no creation arguments
	 #  have two inlets
	 #  (left inlet is what to send,
	 #   right inlet takes ONLY SYMBOLS (or lists) and is the recipient name
	 #   (recipient name initialized to "") )
	 if (length $name == 0) {
	 print qq[
/* object $number is a SEND object with no args */
void object$number (int inletnumber, const struct PDdata * message) {
  char * name = NULL;

  if (message->type == PDC_DSP_REQUEST)
    return;

  switch(inletnumber) {
    case 0:
      sendbyname((name)?name:"",message);
      return;
    case 1:
      if (message->type != PDC_TYPE_SYMBOL) {
        error("(#$number) obj $number: send can only be given a symbol as a recipient (not \%d)",message->type);
        return;
      }
      name = realloc(name,strlen(message->val_symbol));
      if (name == NULL) {
        error("(#$number) obj $number: Memory realloc failed for name (probably fatal, trying anyway)");
        return;
      }
      memcpy(name,message->val_symbol,strlen(message->val_symbol));
      return;
  }
}
];
         } else {
	 # one creation argument
	 #  has one inlet
	 #  (is what to send, recipient is set by creation arg)
	 print qq[
/* object $number is a SEND object with name */
void object$number (int inletnumber, const struct PDdata * message) {

  if (message->type == PDC_DSP_REQUEST || inletnumber != 0)
    return;

  sendbyname("$name",message);
  return;
}
];
       }
       }
	       },
     'receive' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::receive';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 my $name = $arguments[0];

	 # receive, on the other hand, receives from the zero-length name if given no arguments
	 # and can't be changed.

	 # hack to avoid copying
	 $attachedobjectcalls =~ s/&outvalue/message/g;

	 print qq[
/* object $number is a RECEIVE object */
void object$number (int inletnumber, const struct PDdata * message) {
  if (inletnumber != PDC_SENT ||
      message->type == PDC_DSP_REQUEST) {
    return;
  }

  $attachedobjectcalls
}
];

	 $globalinit .= qq[
  /* receive $number*/
  listens(\& object${number},"$name");
];
       }
		  },
     'soundfiler' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::soundfiler';
	 my ($number, %myself, @error) = @_;
	 my $attachedobjectcalls = $myself{AOCS};

	 print qq[
/* object $number is a SOUNDFILER object */
void object$number (int inletnumber, const struct PDdata * message) {
  struct PDdata totable, outvalue;
  int idx, count, bufidx;
  char * filename, * tablename;
  int16_t buf[1024];
  FILE * FH;

  if (inletnumber != 0 || message->type != PDC_TYPE_SYMBOL)
    return;

  if (strcasecmp(message->val_symbol,"read")==0) {
    /* syntax:  read [flags] [filename] [tablename 1] [tablename 2] ... [tablename n]
       i may want to use getopt
       documented flags:
        -skip n (skips first n frames)
        -nframes n (reads not more than n frames)
        -resize (resizes tables to size of file, unless:)
        -maxsize n (do not resize beyond n frames)
        -raw skip chn bps endian (skip skip bytes, read chn channels, each sample
           is bps bytes, endian is one of "l","b","n" to mean "little" "big" or "native" endian

       For my desperately close deadline, I'm going to require that you use "-raw ### 1 2 n"
    */
    if (linkedlength(message)!=8 ||
        strcasecmp(linkedelt(message,1)->val_symbol,"-raw") != 0 ||
                   linkedelt(message,3)->val_float != 1 ||
                   linkedelt(message,4)->val_float != 2 ||
        strcasecmp(linkedelt(message,5)->val_symbol,"n") != 0) {
      error("(#$number) Crippled version of soundfiler; must be invoked as 'read -raw number 1 2 n file table'");
      return;
    }
    filename = linkedelt(message,6)->val_symbol;
    tablename = linkedelt(message,7)->val_symbol;
    totable.val_symbol = "\\\\getguts";
    totable.type = PDC_TYPE_SYMBOL;
    sendbyname(tablename,&totable);
    FH = fopen(filename,"r");
    if (FH == NULL) {
      error("(#$number) Failed to open file: \%s for soundfiler read",filename);
      return;
    }
    fseek(FH,linkedelt(message,2)->val_float,SEEK_SET); idx=0;
    while (!feof(FH) && idx<totable.sample_count) {
      count=fread(buf,sizeof(int16_t),1024,FH);
      for (bufidx=0;bufidx<count && idx<totable.sample_count; bufidx++,idx++)
        totable.dsp_stream[idx]=buf[bufidx] / 32768.0;
    }
    fclose(FH);
  } else if (strcasecmp(message->val_symbol,"write")==0) {
    error("(#$number) soundfiler write unimplemented, sorry.");
  } else error("(#$number) soundfiler only understands symbols 'read' and 'write', not '\%s'",message->val_symbol);
}
];
       }
		     },
		 '\msg' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::\msg';
	 my ($number, %myself, @error) = @_;
	 my $attachedobjectcalls = $myself{AOCS};
	 my @args = @{$myself{ARGS}};

	 # msg syntax looks like:
	 # #X msg x y args

	 print STDERR "PuDaC's message implementation is more tolerant than PureData's!\n"
	   unless ($MSGwarning++);

	 # args may optionally have a symbol as first elt specifying the type of the 2nd elt.
	 # (known types: bang float symbol pointer list)
	 #  BANG: takes no args. resultant message is just a bang.
	 #  FLOAT: takes one arg. resultant message is just the following fp#. If it's not a fpn, error, don't propagate
	 #  SYMBOL: takes one arg. resultant message is just the following symbol. If the symbol is a number, it's truncated to null.
	 #  POINTER: don't exist in my magical little world.
	 #  LIST: takes all args. Casts them all. Used in PD to make things not anythings, but I don't support anythings.

	 # Also: Must interpolate $1 .. $n (note: $0 always is 0 in messages!)

	 # commas separate a single msg box into multiple messages to be sent (in order)
	 # semicolons make the remainder of the message a send
	 # source implies dollars are special? how?

	 # One: See what the first arg is. If one of the above types, cast the remainder appropriately.
	 # Two: For each arg, determine its type (float or symbol)
	 # Three: split on all semicolons and commas. Each comma means restart parsing
	 #  (each semicolon means "instead of propagating, send to the following listener")
	 #  commas immediately after semicolons are ignored -- there's no way to say "now that i've sent
	 #   that message to that listener, i'd like to propagate."
	 #  there's also no way to specify the null listner

	 # XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX
	 # THIS IMPLEMENTATION IS NOT BUG-COMPATIBLE WITH PUREDATA's.
	 # I lump float and symbol together. I pretend anythings are lists, always.
	 # (e.g. [symbol 5< would result in receivers getting "symbol (null-length-symbol)"
	 #  while [float notafloat< would result in an error.
	 # We eat the descriptors at the beginning of messages, so we
	 #  will successfully propagate when PD wouldn't.
	 # I DO NOT SUPPORT SET messages TO MSG BOXES
	 # XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX

	 # So: build original args, resplit on semicolons
	 my @rcvrs = split /;/, (join " ", @args);
	 # lose any initial or terminal commas
	 s/(^\s*,|,\s*$)//g for @rcvrs;

	 # (rcvrs right now contains something like
	 #  "this message gets sent via a wire"
	 #  "this message gets sent to listeners for 'this', as does this one"
	 #  "while this message gets sent to listeners for 'while', and this one" )

	 # separate these into messages to be sent, and messages to propagate

	 # (these are easy)
	 # separate into multiple messages (formerly comma separated)
	 @propagaters = split /,/, (shift @rcvrs);
	 # For each space-separated thing, split it, and put it back
	 for (@propagaters) {
	   s/(^\s*|\s*$)//g;
	   my @t = split / /,$_;
	   $_ = \@t;
	 }
	 # now propagaters should be something like (["this","message"],["that","one"])
	 # now parse type specifiers
	 # (start with "symbol" "float" "list" "bang" (ignore others))

	 for (@propagaters) {
	   if (lc $_->[0] eq "symbol" or
	       lc $_->[0] eq "float") {
	     $_ = [ $_->[1] ];
	   } elsif (lc $_->[0] eq "bang") {
	     $_ = [ $_->[0] ];
	   } elsif (lc $_->[0] eq "list") {
	     shift @$_;
	     # empty lists are bangers.
	     $_ = [ "bang" ]
	       unless (scalar @{$_});
	   }
	 }

	 # generate the dollar-sign parsing code as necessary
	 # also, eat any initial type specifiers

	 # rcvrs looks like:
	 # now it looks like ("one message, two message","red message, blue message")
	 # Next step: split on commas, add the first word of the preceeding message to each.
	 for (my $i=scalar @rcvrs - 1;$i>=0;$i--) {
	   my @expanded = split /,/, $rcvrs[$i];
	   #this should look like ("one message","two message")
	   for (@expanded) {
	     s/(^\s*|\s*$)//g;
	     my @t = split / /,$_;
	     $_ = \@t;
	   }
	   # now this should look like (["one","message"],["two","message"])
	   for (my $j=1;$j<scalar @expanded;$j++) {
	     unshift @{$expanded[$j]}, $expanded[0][0];
	   }
	   # this should now look like (["one","message"],["one","two","message"])
	   $rcvrs[$i] = \@expanded;
	 }
	 # and now (["one","message"],["one","two","message"],["red","message"],["red","blue","message"])

	 # pop off the outer layer of array indirection.
	 @rcvrs = map { @{$_} } @rcvrs;

	 # Also, get rid of any which are just a name: we can't send null messages.
	 @rcvrs = map { $_ if (scalar @$_ >= 2) } @rcvrs;
	 # (technically, PD errors if a receiver doesn't exist, but doesn't send
	 #  a null message if they do)

	 for (@rcvrs) {
	   if (lc $_->[1] eq "symbol" or
	       lc $_->[1] eq "float") {
	     $_ = [ $_->[0] , $_->[2] ];
	   } elsif (lc $_->[0] eq "bang") {
	     $_ = [ $_->[0] , $_->[1] ];
	   } elsif (lc $_->[0] eq "list") {
	     splice @$_,1,1; # get rid of the word "list"
	   }
	 }

	 # TO BE IMPLEMENTED:
	 # More stupid dolsym crap:
	 #  If we're sent a list, and the first elt is not a number
	 #  this may have been an Anything. then we're off by one... augh.
	 # XXXX THIS IS A BUG XXXX

	 # now we have two arrays of arrays (propagaters and rcvrs)
	 # now we need to start cooking these down into native C structures:
	 # { TYPE , floatval , symbolval , 0 , 0 , NULL, { repeat } }
	 my $declstructures="";
	 my $dollarsigncode="";
	 for (my $i=0;$i < scalar @propagaters; $i++) {
	   for (my $j =scalar @{$propagaters[$i]} - 1; $j >= 0; $j--) {
	     $declstructures .= "struct PDdata prop$i".(($j==0)?"":"_$j")." = ";
	     if (lc $propagaters[$i][$j] eq "bang") {
	       # hey, a bang:
	       $declstructures .= "{PDC_TYPE_BANG, 0,NULL,0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&prop${i}_".($j+1)))."};\n";
	     } elsif ($propagaters[$i][$j] =~ /^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) {
	       # looks like a float:
	       $declstructures .= "{PDC_TYPE_FLOAT, $propagaters[$i][$j], NULL,0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&prop${i}_".($j+1)))."};\n";
	     } else {
	       if ($propagaters[$i][$j] =~ /^\$(\d+)$/) {
		 # dolsym - the easy kind: (copy the referrer)
		 $dollarsigncode .= qq[
if (linkedlength(message) < $1) {
  error("(#$number) Too few args for $\%d (had \%d)",$1,linkedlength(message));
} else {
  t = linkedelt(&prop$i,].($j  ).qq[);
  u = linkedelt(message,].($1-1).qq[);
  t->type = u->type;
  if (t->type == PDC_TYPE_FLOAT) t->val_float = u->val_float;
  else if (t->type == PDC_TYPE_SYMBOL) t->val_symbol = u->val_symbol;
}
];
	       } elsif ($propagaters[$i][$j] =~ /\$(\d+)/) {
		 # pain in the neck dolsym: requires interpolation
		 # we can have a$1b
		 # which, if sent "13", should produce a13b
		 # and if sent "symbol bsor" would produce absorb
		 # ... this requires dynamic allocation of memory, EW.
		 croak "Don't yet deal with interpolating messages (specifically".
		   " $propagaters[$i][$j])\n[".(join " ", @args)."<";
	       }
	       # must be a symbol, do symbol (dolsyms are a subset)
	       $declstructures .= "{PDC_TYPE_SYMBOL, 0, \"$propagaters[$i][$j]\",0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&prop${i}_".($j+1)))."};\n";
	     }
	   }
	 }

	 # Now do the same for sent messages
	 for (my $i=0;$i < scalar @rcvrs; $i++) {
	   for (my $j= scalar @{$rcvrs[$i]}-1; $j >= 1; $j--) {
	     $declstructures .= "struct PDdata send$i".(($j==1)?"":"_$j")." = ";
	     if (lc $rcvrs[$i][$j] eq "bang") {
	       # hey, a bang:
	       $declstructures .= "{PDC_TYPE_BANG, 0,NULL,0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&send${i}_".($j+1)))."};\n";
	     } elsif ($rcvrs[$i][$j] =~ /^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) {
	       # looks like a float:
	       $declstructures .= "{PDC_TYPE_FLOAT, $rcvrs[$i][$j], NULL,0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&send${i}_".($j+1)))."};\n";
	     } else {
	       if ($rcvrs[$i][$j] =~ /^\$(\d+)$/) {
		 # dolsym - the easy kind: (copy the referrer)

		 # yes, the $j-1 is supposed to be different here.
		 $dollarsigncode .= qq[
if (linkedlength(message) < $1) {
  error("(#$number) Too few args for $\%d (had \%d)",$1,linkedlength(message));
} else {
  t = linkedelt(&send$i,].($j-1).qq[);
  u = linkedelt(message,].($1-1).qq[);
  t->type = u->type;
  if (t->type == PDC_TYPE_FLOAT) t->val_float = u->val_float;
  else if (t->type == PDC_TYPE_SYMBOL) t->val_symbol = u->val_symbol;
}
];
	       } elsif ($rcvrs[$i][$j] =~ /\$(\d+)/) {
		 # pain in the neck dolsym: requires interpolation
		 # we can have a$1b
		 # which, if sent "13", should produce a13b
		 # and if sent "symbol bsor" would produce absorb
		 # ... this requires dynamic allocation of memory, EW.
		 croak "Don't yet deal with interpolating messages (specifically $rcrs[$i][$j])\n[".
		   (join " ", @args)."<";
	       }
	       # must be a symbol:
	       $declstructures .= "{PDC_TYPE_SYMBOL, 0, \"$rcvrs[$i][$j]\",0,0,NULL,".
		 (($j == scalar @{$propagaters[$i]}-1)?"NULL":("&send${i}_".($j+1)))."};\n";
	     }
	   }
	 }

	 print qq[
/* object $number is a MESSAGE */
void object$number (int inletnumber, const struct PDdata * message) {
  struct PDdata *t;
  const struct PDdata * u;
$declstructures

  /* messages only have one inlet, and are never in DSP chains */
  if (inletnumber != 0 || message->type == PDC_DSP_REQUEST) return;

$dollarsigncode

];
	 # now fix calldowns:
	 for (my $i=0; $i < scalar @propagaters; $i++) {
	   my $t = $attachedobjectcalls;
	   $t  =~ s/outvalue/prop$i/g;
	   print $t;
	 }
	 # now fix sent messages:
	 for (my $i=0; $i < scalar @rcvrs; $i++) {
	   # this is not quite right, as we can interpolate receiver names. Fix it later.
	   print "sendbyname(\"$rcvrs[$i][0]\",&send$i);\n";
	 }

	 # close the subroutine:

	 print "\n}\n";

	 # i guess that's it for \msg.
       }
			    },
     '-' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::-';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a MINUS object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) - takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft - lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft - lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) - doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) - takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) - only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '/' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::/';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a DIVIDE object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) / takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft / lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft / lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) / doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) / takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) / only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '*' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::*';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a MULTIPLY object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) * takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft * lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft * lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) * doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) * takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
        case -1: // No audio
            break;
	default: /* not an inlet */
	    error("(#$number) * only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '>' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::>';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a GREATER-THAN object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) > takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft > lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft > lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) > doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) > takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) > only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '<' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::<';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a LESS-THAN object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) < takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft < lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft < lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) < doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) < takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) < only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     'realtime' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::realtime';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	print qq[
/* object $number is a REALTIME object */
struct timeval object${number}_time;

void object$number (int inletnumber, const struct PDdata * message) {
    struct timeval t;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the COLD inlet (resets count) */
            if (message->type != PDC_TYPE_BANG) {
                error("(#$number) realtime only understands bangs (got \%d)",message->type);
                return;
            }
            gettimeofday(&object${number}_time,NULL);
            return;
	case 1: /* the HOT inlet (sends current time delta) */
            if (message->type != PDC_TYPE_BANG) {
	        error("(#$number) timer only understands bangs");
	        return;
            }
            gettimeofday(&t,NULL);
            outvalue.type = PDC_TYPE_FLOAT;
            outvalue.val_float = (t.tv_sec - object${number}_time.tv_sec)*1000.0;
            outvalue.val_float += (t.tv_usec - object${number}_time.tv_usec)/1000.0;
	    $attachedobjectcalls
	    return;
	default: /* not an inlet */
            return;
    }
}
];

	$globalinit .= qq[
  /* now initializing realtime $number: */
  gettimeofday(&object${number}_time,NULL);
];
}
	  },
     'spigot' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::spigot';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	$attachedobjectcalls =~ s/&outvalue/message/g;

	print qq[
/* object $number is a SPIGOT object */
void object$number (int inletnumber, const struct PDdata * message) {
    static int open = $initializer;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            if (open) {
                $attachedobjectcalls
            }
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) spigot takes only takes float in right inlet");
	        return;
            }
	    open = (message->val_float!=0);
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     'clip' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::clip';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my ($initmin, $initmax) = (0,0);
	$initmin = $arguments[0]
	  if (exists $arguments[0]);
	$initmax = $arguments[1]
	  if (exists $arguments[1]);

	print qq[
/* object $number is a CLIP object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float min = $initmin, max = $initmax, lastout=0;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch (message->type) {
                case PDC_TYPE_BANG:
                    outvalue.type = PDC_TYPE_FLOAT;
                    outvalue.val_float = lastout;
                    $attachedobjectcalls
                    return;
                case PDC_TYPE_FLOAT:
                    lastout = message->val_float;
                    if (message->linked && message->linked->type == PDC_TYPE_FLOAT) {
                        min = message->linked->val_float;
                        if (message->linked->linked && message->linked->linked->type == PDC_TYPE_FLOAT)
                            max = message->linked->linked->val_float;
                    }
                    if (lastout < min) lastout=min;
                    else if (lastout > max) lastout=max;
                    outvalue.type = PDC_TYPE_FLOAT;
                    outvalue.val_float = lastout;
                    $attachedobjectcalls;
                    return;
                default:
                    error("(#$number) clip takes only floats");
                    return;
            }
	    break;
	case 1: /* the cold min inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) spigot takes only takes floats");
	        return;
            }
	    min = message->val_float;
	    break;
	case 2: /* the cold max inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) spigot takes only takes floats");
	        return;
            }
	    max = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     # output for trigger varies. Guess it needs to be a subroutine.
     'trigger' => { 'OT' => [], 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::trigger';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# parse ARGS to some subset of f\S* b\S* s\S* l\S* a\S*
	# i'll just generate all of them and make aliases. (especially since lists and anythings will just be copies)

	# 1- static bang elt
	# 2- semistatic float elt
	# 3- semistatic symbol elt

	for (reverse 0 .. $#arguments) {
	  # have to do the last one first; i was stupid and defined the 1st
	  # output as "outvalue" while subsequent ones are labeled outvalue2, 3, &c.
	  my $curname = "outvalue".(($_!=0)?(1+$_):"");
	  if ($arguments[$_] =~ /^b/i) {
	    $attachedobjectcalls =~ s/$curname/outbang/g;
	  } elsif ($arguments[$_] =~ /^s/i) {
	    $attachedobjectcalls =~ s/$curname/outsym/g;
	  } elsif ($arguments[$_] =~ /^[la]/i) {
	    $attachedobjectcalls =~ s/\&$curname/message/g;
	  } else { # apparently it's otherwise treated as a float
	    $attachedobjectcalls =~ s/$curname/outfloat/g;
	  }
	}

	print qq[
/* object $number is a TRIGGER object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outbang={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};
    struct PDdata outfloat={PDC_TYPE_FLOAT,0,NULL,0,0,NULL,NULL};
    struct PDdata outsym={PDC_TYPE_SYMBOL,0,NULL,0,0,NULL,NULL};
    const char flstr[] = {"float"}, symstr[]={"symbol"};

    if (inletnumber != 0 || message->type == PDC_DSP_REQUEST)
        return;

    switch(message->type) { /* This is right! */
        case PDC_TYPE_FLOAT:
            outfloat.val_float = message->val_float;
            outsym.val_symbol = flstr;
            break;
        case PDC_TYPE_BANG:
            outsym.val_symbol = symstr;
            break;
    }

    /* PuDaC should have renamed everything in the following calls */
    $attachedobjectcalls
    return;
}
];
}
	  },
     'bng' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::bng';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# lots of stupid arguments
	# relevant bit:
	#  if $arguments[3], register for a loadbang
	#  if $arguments[4] != "empty", add a "sendbyname"
	#  if $arguments[5] != "empty", listens for $arguments[5]
	# plus, bng (unlike tgl or friends) doesn't care about [set X< messages

	# our implementation is quite incomplete; there are a wide variety
	# of settings that bangers and toggles are supposed to support (via
	# symbol-value messages)
	# the important missing part is that WE DO NOT CHANGE THE SENDER
	# NAME and CANNOT CHANGE THE RECEIVER NAME. (My architecture was
	# perhaps misdesigned...)

	print qq[
/* object $number is a BANGER */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (!(inletnumber == 0 || inletnumber == PDC_CALLBACK)) return;
    if (message && message->type == PDC_DSP_REQUEST) return;

    $attachedobjectcalls
];
	if ($arguments[4] and lc $arguments[4] ne "empty") {
	  print qq[sendbyname("$arguments[4]",message);\n];
        }
        print qq[
}
];
	if ($arguments[5] and lc $arguments[5] ne "empty") {
	  $globalinit .= qq[
  /* object $number is a banger */
  listens(\& object${number},"$arguments[5]");
];
        }
        if ($arguments[3]) {
	    $loadbangs .= qq[
/* now call banger obj $number: */
object$number(PDC_CALLBACK,NULL);
];
        }

 }
          },
     'tgl' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::tgl';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# lots of stupid arguments
	# relevant bit:
	#  if $arguments[1], set on creation
	#  if $arguments[2] != "empty", add a "sendbyname"
	#  if $arguments[3] != "empty", listens for $arguments[5]
	#  $arguments[12] contains the non-zero value if $arguments[1]
	#  $arguments[13] contains the non-zero value if not $arguments[1]

	my ($nzval,$curval)=(1,0);
        if ($arguments[1] and $arguments[12]) {
	  $nzval = $curval = $arguments[12];
	} else {
	  $nzval = $arguments[13];
        }

	# our implementation is quite incomplete; there are a wide variety
	# of settings that bangers and toggles are supposed to support (via
	# symbol-value messages)

	# the important missing part is that WE DO NOT CHANGE THE SENDER
	# NAME and CANNOT CHANGE THE RECEIVER NAME. (My architecture was
	# perhaps misdesigned...)

	print qq[
/* object $number is a TOGGLE */
void object$number (int inletnumber, const struct PDdata * message) {
    static float nonzerovalue=$nzval, curvalue=$curval;
    struct PDdata outvalue={PDC_TYPE_FLOAT,0,NULL,0,0,NULL,NULL};

    if (message->type == PDC_DSP_REQUEST) return;
    if (!(inletnumber == 0 || inletnumber == PDC_CALLBACK || inletnumber == PDC_SENT)) return;

    switch (message->type) {
        case PDC_TYPE_FLOAT:
            if (message->val_float == 0)
                curvalue =0;
            else
                nonzerovalue = curvalue = message->val_float;
            break;
        case PDC_TYPE_BANG:
            if (curvalue == 0) curvalue = nonzerovalue;
            else curvalue = 0 ;
            break;
        case PDC_TYPE_SYMBOL:
            if (strcasecmp("set",message->val_symbol)==0) {
                if (message->linked && message->linked->type == PDC_TYPE_FLOAT) {
                    if (message->linked->val_float == 0) curvalue = 0;
                    else nonzerovalue = curvalue = message->linked->val_float;
                } else
                    error("(#$number) toggle box doesn't understand message 'set [nonfloat]'");
            } else
                error("(#$number) toggle box doesn't understand message starting with \%s",message->val_symbol);
            return;
    }

    outvalue.val_float = curvalue;
    $attachedobjectcalls
];
	if ($arguments[2] and lc $arguments[2] ne "empty") {
	  print qq[sendbyname("$arguments[2]",message);\n];
        }
        print qq[
}
];
	if ($arguments[3] and lc $arguments[3] ne "empty") {
	  $globalinit .= qq[
  /* object $number is a toggle */
  listens(\& object${number},"$arguments[5]");
];
        }
 }
          },
     '\floatatom' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::\floatatom';
	my ($number, %myself, @error) = @_;
	my ($rcv,$snd) = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# lots of stupid arguments (already stripped in xtypes)
	# relevant bit:
	#  if $snd != "-", add a "sendbyname"
	#  if $rcv != "-", listens for $rcv

	print qq[
/* object $number is a FLOATATOM */
void object$number (int inletnumber, const struct PDdata * message) {
    static float curvalue=0;
    struct PDdata outvalue={PDC_TYPE_FLOAT,0,NULL,0,0,NULL,NULL};

    if (message->type == PDC_DSP_REQUEST) return;
    if (!(inletnumber == 0 || inletnumber == PDC_CALLBACK || inletnumber == PDC_SENT)) return;

    switch (message->type) {
        case PDC_TYPE_FLOAT:
            curvalue = message->val_float;
            break;
        case PDC_TYPE_SYMBOL:
            if (strcasecmp("set",message->val_symbol)==0) {
                if (message->linked && message->linked->type == PDC_TYPE_FLOAT) {
                    curvalue = message->linked->val_float;
                } else
                    error("(#$number) floatatom doesn't understand message 'set [nonfloat]'");
            } else
                error("(#$number) floatatom doesn't understand message starting with \%s",message->val_symbol);
            return;
    }

    outvalue.val_float = curvalue;
    $attachedobjectcalls
/*}*/]; #]
	if ($arguments[2] and lc $arguments[2] ne "-") {
	  print qq[sendbyname("$arguments[2]",message);\n];
        }
        print "}\n";

	if ($arguments[3] and lc $arguments[3] ne "-") {
	  $globalinit .= qq[
  /* object $number is a floatatom */
  listens(\& object${number},"$arguments[5]");
]; #]
        }
 }
          },
     'unpack' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::unpack';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# unpacks have # args = # outlets (except 0 args = "f f")
	# all args are s\S* f\S* \d+ p\S*
	# it's either explicitly a Symbol, a Pointer, or else it's a float.

	# so handle default:
	@arguments=("f","f") if (scalar @arguments == 0);

	# unbundle attachedobjectcalls:
	my @AOCsByOutlet;
	for (split "\n",$attachedobjectcalls) {
	  if (/outvalue(\d*)/) {
	    my $t = $1;
	    if ($t>0) { $t--; }
	    else { $t=0; }
	    if ($arguments[$t] =~ /^s/i) { # symbol
	      s/outvalue\d*/outsymb/;
	    } else {
	      s/outvalue\d*/outfloat/;
	    }
	    $AOCsByOutlet[$t] .= "$_\n";
	  } else {
	    carp "Didn't understand a call i generated?  [$_]\n";
	  }
	}

	print qq[
/* object $number is an UNPACK object */
void object$number (int inletnumber, const struct PDdata * message) {
    int i;
    struct PDdata outfloat={PDC_TYPE_FLOAT,0,NULL,0,0,NULL,NULL};
    struct PDdata outsymb={PDC_TYPE_SYMBOL,0,NULL,0,0,NULL,NULL};

    if (message->type == PDC_DSP_REQUEST || !(inletnumber == 0)) return;
/*} cperl */
];#]
	for (my $j=$#arguments;$j>=0;$j--) {
          next if (length $AOCsByOutlet[$j]==0);
	  if ($arguments[$j] =~ /^s/i) {
	    # a symbol:
	    print qq[
    if (linkedlength(message) > $j && linkedelt(message,$j)->type == PDC_TYPE_SYMBOL) {
        outsymb.val_symbol = linkedelt(message,$j)->val_symbol;
        $AOCsByOutlet[$j];
    } else
        error("(#$number) Unpack got non-symbol for a symbol slot");
];#]
          } elsif ($arguments[$i] =~ /^p/i) {
	    carp "PuDaC doesn't deal with pointers";
	  } else { # a float
	    print qq[
    if (linkedlength(message) > $j && linkedelt(message,$j)->type == PDC_TYPE_FLOAT) {
        outfloat.val_float = linkedelt(message,$j)->val_float;
        $AOCsByOutlet[$j];
    } else
        error("(#$number) Unpack got non-symbol for a symbol slot");
]; #]
	  }
	}

	print "}\n";
}
          },
     'select' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::select';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# select comes in four flavors: float selector and symbol selector; zero-or-one args or more
	# augh!

	# select with no args compares things to zero.

	# if the first arg is a float, we cast everything else to float (symbols become 0)

	@arguments = ( 0 )
	  if (scalar @arguments == 0);

	# unbundle attachedobjectcalls:
	my @AOCsByOutlet;
	for (split "\n",$attachedobjectcalls) {
	  if (/outvalue(\d*)/) {
	    my $t = $1;
	    if ($t>0) { $t--; }
	    else { $t=0; }
	    if ($t != scalar @arguments) { # symbol
	      s/outvalue\d*/outvalue/;
	    } else {
	      s/\&outvalue\d*/message/;
	    }
	    $AOCsByOutlet[$t] .= "$_\n";
	  } else {
	    carp "Didn't understand a call i generated?  [$_]\n";
	  }
	}

	if (scalar @arguments <= 1) {
	  my $init = $arguments[0];
	  if ($arguments[0] =~ /^-?(\d*\.)?\d+([eE][-+]?\d+)?$/ or not $arguments[0]) {
	    # is no args or a single numeric arg:
	    print qq[
/* object $number is a SELECT object with 0/1 numeric args */
void object$number (int inletnumber, const struct PDdata * message) {
    static float comvalue=$init;
    struct PDdata outvalue={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (!(message->type == PDC_TYPE_FLOAT || message->type == PDC_TYPE_SYMBOL))
      return;

    switch (inletnumber) {
        case 0: /* HOT */
            if (message->type == PDC_TYPE_FLOAT && comvalue == message->val_float) {
                $AOCsByOutlet[0]
            } else {
                $AOCsByOutlet[1]
            }
            break;
        case 1: /* COLD */
            if (message->type != PDC_TYPE_FLOAT) {
                error("(#$number) This selector ($number) only knows how to compare floats");
                return;
	    }
            comvalue = message->val_float;
            break;
    }
}
];
	  } else {
	    # is 1 non-numeric arg:
	    print qq[
/* object $number is a SELECT object with 1 symbol arg */
void object$number (int inletnumber, const struct PDdata * message) {
    static char * comvalue = NULL;
    struct PDdata outvalue={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (comvalue == NULL) {
        comvalue = malloc(1+strlen("$init"));
        if (comvalue == NULL) { error("(#$number) MEMORY ALLOC FAIL"); exit(1); }
        strcpy(comvalue,"$init");
    }

    if (!(message->type == PDC_TYPE_FLOAT || message->type == PDC_TYPE_SYMBOL))
      return;

    if (message->type == PDC_DSP_REQUEST) return;
    switch (inletnumber) {
        case 0: /* HOT */
            if (message->type == PDC_TYPE_SYMBOL && strcmp(comvalue,message->val_symbol)==0) {
                $AOCsByOutlet[0]
            } else {
                $AOCsByOutlet[1]
            }
            break;
        case 1: /* COLD */
            if (message->type != PDC_TYPE_SYMBOL) {
                error("(#$number) This selector ($number) only knows how to compare symbols");
                return;
	    }
            comvalue = realloc(comvalue,1+strlen(message->val_symbol));
            if (comvalue == NULL) { error("(#$number) MEMORY ALLOC FAIL"); exit(1); }
            strcpy(comvalue,message->val_symbol);
            break;
      }
}
];
	  }
        } else { # has more than one arg, ugh
	  if ($arguments[0] =~ /^-?(\d*\.)?\d+([eE][-+]?\d+)?$/ or not $arguments[0]) {
	    # is comparing any number of floats
	    # first force all args to float type
	    @arguments = map { (/^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) ? $_ : 0 } @arguments;
	    # fortunately we can't change anything after creation.
	    print qq[
/* object $number is a SELECT object with >= 2 numeric args */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (!(message->type == PDC_TYPE_FLOAT || message->type == PDC_TYPE_SYMBOL) || inletnumber != 0 )
      return;
];
	    if (length $AOCsByOutlet[scalar @arguments]) {
	      print qq[
    if (message->type == PDC_TYPE_SYMBOL) {
        ].$AOCsByOutlet[scalar @arguments].qq[
        return;
    }
];
	    }

	    for (my $i=0;$i<=$#arguments;$i++) {
	      # apparently PD itself just does == on the floating point numbers?
	      print qq[
    if (message->val_float == $arguments[$i]) {
        $AOCsByOutlet[$i]
        return;
    }
];
	    }
	    print "\n".$AOCsByOutlet[scalar @arguments]."\n}\n";
	  } else {
	    # is comparing any number of symbols
	    # first convert all fp numbers to null strings. (augh, stupid puredata)
	    @arguments = map { (/^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) ? "" : $_ } @arguments;

	    print qq[
/* object $number is a SELECT object with >= 2 symbolic args */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (!(message->type == PDC_TYPE_FLOAT || message->type == PDC_TYPE_SYMBOL) || inletnumber != 0 )
      return;
];
	    if (length $AOCsByOutlet[scalar @arguments]) {
	      print qq[
    if (message->type == PDC_TYPE_FLOAT) {
        ].$AOCsByOutlet[scalar @arguments].qq[
        return;
    }
];
	    }

	    for (my $i=0;$i<=$#arguments;$i++) {
	      # apparently PD itself just does == on the floating point numbers?
	      print qq[
    if (strcmp(message->val_symbol,"$arguments[$i]")==0) {
        $AOCsByOutlet[$i]
        return;
    }
];
	    }
	    print "\n".$AOCsByOutlet[scalar @arguments]."\n}\n";
	  }
	}
}
          },
     'route' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::route';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# route comes in two flavors: float router, and symbol/type router

	# symbol/type router is technically a side affect of how puredata
	# stores types (with a word at the beginning specifying
	# type. oops.)

	# route without any args compares things to zero

	# if the first arg is a float, we cast everything else to float (symbols become 0)

	@arguments = ( 0 )
	  if (scalar @arguments == 0);

	print STDERR "PuDaC doesn't have Anythings, so routing by type is Very different\n"
	  unless ($ROUTEwarning++);

	# unbundle attachedobjectcalls:
	my @AOCsByOutlet;
	for (split "\n",$attachedobjectcalls) {
	  if (/outvalue(\d*)/) {
	    my $t = $1;
	    if ($t>0) { $t--; }
	    else { $t=0; }
	    if ($t != scalar @arguments) { # symbol
	      s/\&outvalue\d*/message->linked/g;
	    } else {
	      s/\&outvalue\d*/message/g;
	    }
	    $AOCsByOutlet[$t] .= "$_\n";
	  } else {
	    carp "Didn't understand a call i generated?  [$_]\n";
	  }
	}

	if ($arguments[0] =~ /^-?(\d*\.)?\d+([eE][-+]?\d+)?$/ or not $arguments[0]) {
	  # is comparing any number of floats
	  # first force all args to float type
	  @arguments = map { (/^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) ? $_ : 0 } @arguments;
	  # fortunately we can't change anything after creation.
	  print qq[
/* object $number is a ROUTE object with numeric args */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata banger={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (!(message->type == PDC_TYPE_FLOAT || message->type == PDC_TYPE_SYMBOL) || inletnumber != 0 )
      return;
];
	  if (length $AOCsByOutlet[scalar @arguments]) {
	    print qq[
    if (message->type == PDC_TYPE_SYMBOL) {
        ].$AOCsByOutlet[scalar @arguments].qq[
        return;
    }
];
	  }

	  for (my $i=0;$i<=$#arguments;$i++) {
	    # apparently PD itself just does == on the floating point numbers?
	    print qq[
    if (message->val_float == $arguments[$i]) {
        if (message->linked) {
            $AOCsByOutlet[$i]
        } else { ];
	    my $t = $AOCsByOutlet[$i]; $t =~ s/message->linked/&banger/g;
	    print qq[
            $t
        }
        return;
    }
];
	  }
	  print "\n".$AOCsByOutlet[scalar @arguments]."\n}\n";
	} else {
	  # is comparing any number of symbols
	  # first convert all fp numbers to null strings. (augh, stupid puredata)
	  @arguments = map { (/^-?(\d*\.)?\d+([eE][-+]?\d+)?$/) ? "" : $_ } @arguments;
	  print qq[
/* object $number is a ROUTE object with symbolic args / type router */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata banger={PDC_TYPE_BANG,0,NULL,0,0,NULL,NULL};

    if (message->type == PDC_DSP_REQUEST || inletnumber != 0 )
      return;
];
	  # XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX
	  # here's another place where my abstraction's gotten me into trouble:
	  # I can't tell the difference between symbols and anythings.
	  # To work around this, we make symbol and list bind more loosely than anything but failure.

	  # maybe I should make a hash?
	  for (my $i=$#arguments;$i>=0;$i--) {
	    $codefor{$arguments[$i]} = $AOCsByOutlet[$i]; # the assignment occludes later matches
	  }

	  if (exists $codefor{'bang'}) {
	    $codefor{'bang'} =~ s/message->linked/message/g;
	    print qq[
    if (message->type == PDC_TYPE_BANG) {
        $codefor{bang}
        return;
    }
];
	  }
	  if (exists $codefor{'float'}) {
	    $codefor{'float'} =~ s/message->linked/message/g;
	    print qq[
    if (message->type == PDC_TYPE_FLOAT && message->linked == NULL) {
        $codefor{float}
        return;
    }
];
	  }
	  # now we try everything /except/ bang float symbol and list
	  for (keys %codefor) {
	    # (bang and float we've already tested, so we skip them for non-stupid code. )
	    #  list and symbol should bind more loosely 
	    next if (/^(bang|float|list|symbol)$/);
	    #everything else should be the first elt of an anything
	    print qq[
    if (message->type == PDC_TYPE_SYMBOL && strcmp(message->val_symbol,"$_")==0) {
        $codefor{$_}
        return;
    }
];
          }
	  if (exists $codefor{'symbol'}) {
	    print qq[
    if (message->type == PDC_TYPE_SYMBOL && message->linked == NULL) {
        $codefor{symbol}
        return;
    }
];
	  }
	  if (exists $codefor{'list'}) {
	    print qq[
    if (message->linked != NULL) {
        $codefor{list}
        return;
    }
];
	  }
	  # default:
	  print "\n".$AOCsByOutlet[scalar @arguments]."\n}\n";
	}
}
          },
     'netreceive' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::netreceive';
	my ($number, %myself, @error) = @_;
	my ($port,$mode) = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# port: zero means any; non-zero means that port
	# mode: zero means TCP; non-zero means UDP

	if ($mode) { # UDP
	  $attachedobjectcalls =~ s/\&outvalue/outarray/g;

	  print qq[
/* object $number is a NETRECEIVE (UDP) object */
int object${number}_fd;
struct sockaddr_in object${number}_saddr;
struct pollfd object${number}_poll;

void object$number (int inletnumber, const struct PDdata * message) {
    int n, i, from, to;
    char buffer[4097];
    char *p, *tester, *restart, *nextcomma;
    char * pointers[2048];
    struct PDdata outarray[2048];
    float t;

    n = recv(object${number}_fd,buffer+1,4096,0); // this version blocks. need to fix that.
    // recv doesn't add a null terminator.
    buffer[n+1]=0;
    buffer[0]=',';
    // PD's netreceive:
    //  --UDP version--
    //  Doesn't parse anything after the first semicolon on a line
    //  Each datagram is parsed as a message, regardless of whether it contains a semicolon
    //  Commas divide a message into multiple messages
    //  --TCP version--
    //  Each semicolon separates messages.
    //  Commas divide a message into multiple messages
    //  Packets are accumulated until a semicolon is found

    // Either way, it does the stupid named casting thing that's getting me in trouble everywhere.

    memset(pointers,0,2048*sizeof(char *));
    memset(outarray,0,2048*sizeof(struct PDdata));

    // With UDP datagrams, if it contains a semicolon, we forget everything after.
    if (p = strchr(buffer,';')) {
      *p = 0;
      n=p-buffer;
    }

    restart=buffer;
    while (restart-buffer < n) {
      restart++;
      nextcomma = strchr(restart,',');
      if (nextcomma) *nextcomma = 0;
      else nextcomma = buffer+n;

      // Parse things out.
      p=restart; i=0;
      do {
	p--; //stupid fidgeting
	while (isspace(*++p)) { if (! *p) break; }
	//T should point to next non-space, so make note of it
	pointers[i++]=p;

	while (!isspace(*++p)) { if (! *p) break; }
	// T should point to next space, so null-terminate
	*p++=0; // more stupid fidgeting

      } while (p < nextcomma && i < 2047);

      if (strcmp(pointers[0],"bang")==0) {
	outarray[0].type = PDC_TYPE_BANG;
	from=to=i; // do not parse
      } else if (strcmp(pointers[0],"float")==0) {
	outarray[0].type = PDC_TYPE_FLOAT;
	outarray[0].val_float = strtod(pointers[1],NULL);
	from=to=i;
      } else if (strcmp(pointers[0],"symbol")==0) {
	outarray[0].type = PDC_TYPE_SYMBOL;
	outarray[0].val_symbol = pointers[1];
	from=to=i;
      } else if (strcmp(pointers[0],"list")==0) {
	from=1; to=0;
      } else {
	from=to=0;
      }

      for (;from<i;from++,to++) {
	t = strtod(pointers[from],&tester); // I'd like to use strtof and avoid the cast, but ...
	if (tester-pointers[from] == strlen(pointers[from])) {
	  outarray[to].type = PDC_TYPE_FLOAT;
	  outarray[to].val_float = t;
	} else { // must be a symbol
	  outarray[to].type = PDC_TYPE_SYMBOL;
	  outarray[to].val_symbol = pointers[from];
	}
	outarray[to].linked=&outarray[to+1];
      }
      outarray[to-1].linked=NULL; // undo that.

      $attachedobjectcalls

      restart = nextcomma;
    }
}
];

	 $globalinit .= qq[
  /* now initializing netreceive $number: */
  if((object${number}_fd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))<0) {
    error("(#$number) can't create udp socket, aborting");
    exit(2);
  }

  object${number}_saddr.sin_family      = AF_INET;
  object${number}_saddr.sin_port        = htons($port);
  object${number}_saddr.sin_addr.s_addr = htonl(INADDR_ANY);

  if(bind(object${number}_fd,(struct sockaddr *)&object${number}_saddr,sizeof(struct sockaddr_in))<0) {
    error("(#$number) can't bind udp socket to port $port");
    perror("(#$number) ");
    exit(2);
  }

  object${number}_poll.fd = object${number}_fd;
  object${number}_poll.events = POLLIN;
];
	 $periteration .= qq[
/* Now checking if traffic has arrived for netreceive $number: */
/* This implementation may yield a DoS? Maybe s/while/if/ */
   while (poll(&object${number}_poll,1,0)) {
      object$number(PDC_CALLBACK,NULL);
   }
];
	} else {
	  croak "Haven't implemented TCP netreceive yet, sorry.";
	}
}
	  },
     'tabread~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::tabread~';
	 my ($number, %myself, @error) = @_;
	 my ($name) = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 print qq[
/* object $number is a TABREAD~ object */
float *object${number}_upstr, *object${number}_cache;

void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata totable={PDC_TYPE_SYMBOL,0,"\\\\getguts",0,0,NULL,NULL};
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache = NULL;
    static float * table = NULL;
    static int tablesize=0;
    float * writer;
    static unsigned int pos=0;
    int i;
    char * write_is_dsp;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr;
    if (NULL == cache) cache = object${number}_cache;
    if (NULL == table) {
      sendbyname("$name",&totable);
      table = totable.dsp_stream;
      tablesize = totable.sample_count; /* This is bad. We can't dynamically resize things */
    }

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* find out if we have a dsp rate input */
            if (message->sample_count != DSPtrace1.sample_count) {
                memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
                DSPtrace1.sample_count = message->sample_count;
                $reverseDSPcalls
                if (DSPtrace1.is_dsp) {
                    for (i=0;i<block_size;i++) {
                        if (DSPtrace1.dsp_stream[i] < 0) cache[i] = table[0];
                        else if (DSPtrace1.dsp_stream[i] >= tablesize) cache[i] = table[tablesize-1];
                        else cache[i] = table[(unsigned int)DSPtrace1.dsp_stream[i]];
                    }
                } else {
                    for (i=0;i<block_size;i++) {
                        cache[i]=table[pos];
                    }
                }
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 0: /* inlet 1 = position */
            if (message->type == PDC_TYPE_FLOAT) {
                if (message->val_float < 0) pos = 0;
                else if (message->val_float >= tablesize) pos=tablesize-1;
                else pos = message->val_float;
            } else
                error("(#$number) tabread~ takes only floats or dsp chains in idx inlet (got \%d)",message->type);

            break;
        default:
            error("(#$number) tabread~ only has one inlet (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for tabread~ $number */
  object${number}_upstr = malloc(block_size*sizeof(float));
  object${number}_cache = malloc(block_size*sizeof(float));

  if (object${number}_upstr == NULL || object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (tabread~)!");
    exit(1);
  }
];
       }
	       },
     'line~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::line~';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};

	 print qq[
/* object $number is an LINE~ object */
float *object${number}_cache;

void object$number (int inletnumber, const struct PDdata * message) {
    static int cachepos=-1;
    static float * cache = NULL;
    float * writer;
    static float curpos=0, dest=0;
    static int blockstogo=0;
    int i;
    char * write_is_dsp;

    if (NULL == cache) cache = object${number}_cache;

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* find out if we have a dsp rate input */
            if (message->sample_count != cachepos) {
                memset(cache,0,block_size*sizeof(float));
                cachepos = message->sample_count;
                if (blockstogo) {
                    /* (dest-curpos)/blockstogo */
                    for (i=0;i<block_size;i++)
                        cache[i]=i*((dest-curpos)/blockstogo)/block_size+curpos;
                    curpos += (dest-curpos)/blockstogo;
                    blockstogo--;
                } else
                    for (i=0;i<block_size;i++)
                        cache[i] = dest;
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 0: /* inlet 1 = numbers or lists */
            if (message->type == PDC_TYPE_FLOAT)
                dest = message->val_float;
            else {
                error("(#$number) line~ takes only floats or dsp chains in left inlet (got \%d)",message->type);
                return;
            }

            if (message->linked)
                if (message->linked->type == PDC_TYPE_FLOAT) {
                    blockstogo = message->linked->val_float * sample_rate / block_size / 1000;
                    if (blockstogo < 0) blockstogo = 0;
                } else {
                    error("(#$number) line~ takes only float as 2nd elt of list in 1st inlet (got \%d)",message->type);
                    return;
                }
            else curpos=dest; // Act immediately if we got a single number

            break;
        case 1: /* inlet 2 = time for next line~ to take */
            /* only if numeric */
            error("(#$number) line~ incomplete; does not accept input in right inlet.");
            break;
        default:
            error("(#$number) line~ only has two inlets (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for line~ $number */
  object${number}_cache = calloc(block_size,sizeof(float));

  if (object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (line~)!");
    exit(1);
  }
];
       }
	       },
     'max' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::max';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a MAX object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                if (message->linked != NULL) {
                    error("(#$number) max doesn't understand lists");
                    return;
                }
                lastleft = message->val_float;
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = (lastleft>lastright)?lastleft:lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = (lastleft>lastright)?lastleft:lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) + doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) max takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     'min' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::min';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a MIN object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                if (message->linked != NULL) {
                    error("(#$number) min doesn't understand lists");
                    return;
                }
                lastleft = message->val_float;
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = (lastleft<lastright)?lastleft:lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = (lastleft<lastright)?lastleft:lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) + doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) min takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];
}
	  },
     'sqrt' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::sqrt';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is a SQRT object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue;

    outvalue.linked = NULL;
    if (message->type != PDC_TYPE_FLOAT || inletnumber != 0)
        return;

    outvalue.type = PDC_TYPE_FLOAT;
    if (message->val_float > 0)
        outvalue.val_float = sqrt(message->val_float);
    else
        outvalue.val_float = 0;

    $attachedobjectcalls
}
];
}
	  },
     'abs' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::abs';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	print qq[
/* object $number is an ABS object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue;

    outvalue.linked = NULL;
    if (message->type != PDC_TYPE_FLOAT || inletnumber != 0)
        return;

    outvalue.type = PDC_TYPE_FLOAT;
    outvalue.val_float = fabs(message->val_float);

    $attachedobjectcalls
}
];
}
	  },
     'delay' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::delay';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 # i've evidently decided to implement delay in terms of metro.

	 my $initializer = 0;
	 $initializer = $arguments[0]
	   if (exists $arguments[0]);

	 print qq[
/* object $number is a DELAY object */

struct timeval object${number}_time;
int object${number}_running;

void object$number (int inletnumber, const struct PDdata * message) {
    static float delay_in_milliseconds=$initializer;
    struct PDdata outvalue;
    struct timeval TV_internal;

    outvalue.linked = NULL;
    outvalue.type = PDC_TYPE_BANG;

    switch (inletnumber) {
        case PDC_CALLBACK: /* a timeout */
            object${number}_running=0;
            /* this order is vital: otherwise we can't trigger ourselves to restart*/
            $attachedobjectcalls
            break;
	case 0: /* the hot inlet */
            switch(message->type) {
            case PDC_TYPE_FLOAT: /* N means schedule a timeout for N milliseconds */
                delay_in_milliseconds = message->val_float;
            case PDC_TYPE_BANG: /* bang means run */
                object${number}_running = 1;
                break;
            case PDC_TYPE_SYMBOL: /* "stop" means stop */
                if (strcasecmp("stop",message->val_symbol)==0) {
                    object${number}_running = 0;
                    return;
                }
                error("(#$number) delay doesn't understand that symbol (\%s)",message->val_symbol);
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    if (object${number}_running) {
                gettimeofday(&TV_internal,NULL);
                object${number}_time.tv_usec = TV_internal.tv_usec
                                               + delay_in_milliseconds*1000;
                object${number}_time.tv_sec = TV_internal.tv_sec;
                while (object${number}_time.tv_usec > 1000000) {
                    object${number}_time.tv_sec++;
                    object${number}_time.tv_usec-=1000000;
                }
            }
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) delay takes only takes float in right inlet");
	        return;
            }
	    delay_in_milliseconds = message->val_float;
	    break;
	default: /* not an inlet */
            return;
    }
    return;
}
];

	 $globalinit .= qq[
  /* now initializing delay $number: */
  object${number}_running = 0;
];
	 $periteration .= qq[
/* Now checking timeout for obj $number: */
if (object${number}_running &&
    (now.tv_sec > object${number}_time.tv_sec ||
     (now.tv_sec == object${number}_time.tv_sec &&
      now.tv_usec > object${number}_time.tv_usec))) {
       object${number}(PDC_CALLBACK,NULL);
}
];
       }
		},
     'pack' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::pack';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	# packs have # args = # inlets (except 0 args = "f f")
	# all args are s\S* f\S* \d+ p\S*
	# it's either explicitly a Symbol, a Pointer, or else it's a float.

	# so handle default:
	@arguments=("f","f") if (scalar @arguments == 0);

	s/^f$/0/ig for (@arguments);

	print qq[
/* object $number is an PACK object */
void object$number (int inletnumber, const struct PDdata * message) {
    int i;
    static struct PDdata outvalue[]={];
	for (my $j=0;$j<=$#arguments;$j++) {
	  if ($arguments[$j] =~ /^s/i) { # a symbol
	    print "{PDC_TYPE_SYMBOL,0,\"symbol\",0,0,0,";
	    if ($j==$#arguments) { print "NULL}};\n" }
	    else { print "&outvalue[".($j+1)."]},"; }
	  } elsif ($arguments[$i] =~ /^p/i) {
	    croak "PuDaC doesn't deal with pointers";
	  } else { # a float
	    print "{PDC_TYPE_FLOAT,$arguments[$j],0,0,0,0,";
	    if ($j==$#arguments) { print "NULL}};\n" }
	    else { print "&outvalue[".($j+1)."]},"; }
	  }
	}

	print qq[

    if (message->type == PDC_DSP_REQUEST || inletnumber < 0) return;

    switch(inletnumber) {
];
	for (my $j=0;$j<=$#arguments;$j++) {
	  if ($arguments[$j] =~ /^s/i) { # a symbol:
	    print qq[
    case $j:
        if (message->type == PDC_TYPE_SYMBOL) {
            outvalue[$j].val_symbol = message->val_symbol; // Static-ness of this will probably break!
            error("(#$number) Likely bug in pack. Check your symbols."); ].
            (($j==0)?$attachedobjectcalls:"").qq[
        } else
            error("(#$number) pack got non-symbol for a symbol slot");
        break;
];
          } elsif ($arguments[$i] =~ /^p/i) {
	    carp "PuDaC doesn't deal with pointers";
	  } else { # a float
	    print qq[
    case $j:
        if (message->type == PDC_TYPE_FLOAT) {
            outvalue[$j].val_float = message->val_float; ].
            (($j==0)?$attachedobjectcalls:"").qq[
        } else
            error("(#$number) pack got non-float for a float slot");
        break;
];
	  }
	}

	print "}\n}\n";
}
          },
     'until' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::+';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	print qq[
/* object $number is an UNTIL object */
void object$number (int inletnumber, const struct PDdata * message) {
    static unsigned int counts=-1;
    struct PDdata outvalue={PDC_TYPE_BANG,0,0,0,0,0,0};
    static int counter=0; // must be static so all calls of this until are accessing the same number

    switch (inletnumber) {
	case 0: /* the burning hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                counts = message->val_float;
                break;
            case PDC_TYPE_BANG:
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) until doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
            counter = counts;
            while (counter--) {
                $attachedobjectcalls
            }
	    break;
	case 1: /* the stopping cold inlet */
            if (message->type != PDC_TYPE_BANG) {
	        error("(#$number) + takes only takes float in right inlet");
	        return;
            }
	    counter=0;
	    break;
    }
    return;
}
];
}
	  },
     '==' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::==';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is an EQUALS object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) == takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft == lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft == lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) == doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) == takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) == only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '!=' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::==';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	my $initializer = 0;
	$initializer = $arguments[0]
	  if (exists $arguments[0]);

	print qq[
/* object $number is an NOT-EQUALS object */
void object$number (int inletnumber, const struct PDdata * message) {
    static float lastleft=0, lastright=$initializer;
    struct PDdata outvalue;

    outvalue.linked = NULL;

    switch (inletnumber) {
	case 0: /* the hot inlet */
	    /* here the compiler needs to insert the calls to all attached objects*/
            switch(message->type) {
            case PDC_TYPE_FLOAT:
                lastleft = message->val_float;
                if (message->linked != NULL) {
                    if (message->linked->type != PDC_TYPE_FLOAT) {
                        error("(#$number) != takes only float as 2nd elt in list");
                        /* however this doesn't fail to propagate */
                    } else {
                        lastright = message->linked->val_float;
                    }
                }
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft != lastright;
                break;
            case PDC_TYPE_BANG:
                outvalue.type = PDC_TYPE_FLOAT;
                outvalue.val_float = lastleft != lastright;
                break;
            case PDC_TYPE_SYMBOL:
                error("(#$number) != doesn't understand symbols");
                return;
            default:
                error("(#$number) unknown type \%d",message->type);
                return;
            }
	    $attachedobjectcalls
	    break;
	case 1: /* the cold inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) != takes only takes float in right inlet");
	        return;
            }
	    lastright = message->val_float;
	    break;
	default: /* not an inlet */
	    error("(#$number) != only has two inlets");
            return;
    }
    return;
}
];
}
	  },
     '*~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::*~';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 # Two forms of (binop)~: float*float and float*event
	 #  determined by whether there is a creation argument.

	 if (scalar @arguments) {
	   print qq[
/* object $number is a stream*event *~ object */
float *object${number}_upstr, *object${number}_cache;

void object$number (int inletnumber, const struct PDdata * message) {
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache = NULL;
    float * writer;
    char * write_is_dsp;
    float scalar = $arguments[0];
    int i;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr;
    if (NULL == cache) cache = object${number}_cache;

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* we always have a dsp rate input on left */
            if (message->sample_count != DSPtrace1.sample_count) {
                memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
                DSPtrace1.sample_count = message->sample_count;
                $reverseDSPcalls
                for (i=0;i<block_size;i++) {
                    cache[i]=DSPtrace1.dsp_stream[i]*scalar;
                }
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 1: /* inlet 2 = adjust scalar */
            if (message->type == PDC_TYPE_FLOAT)
                scalar = message->val_float;
            else
                error("(#$number) this *~ takes only floats in scalar inlet (got \%d)",message->type);

            break;
        case 0: /* inlet 1 = invalid */
            error("(#$number) *~ takes only audio in left inlet (got \%d)",message->type);
            break;
        default:
            error("(#$number) *~ only has two inlets (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for *~ $number */
  object${number}_upstr = malloc(block_size*sizeof(float));
  object${number}_cache = malloc(block_size*sizeof(float));

  if (object${number}_upstr == NULL || object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (*~)!");
    exit(1);
  }
];
	 } else {
	   print STDERR "PuDaC does not support all the vagaries of argumentless *~ -- currently only supports stream*stream\n"
	     unless ($PuDaC_multiply_stream_warning++);
	   print qq[
/* object $number is a stream*stream *~ object */
float *object${number}_upstr1, *object${number}_cache, *object${number}_upstr2;

void object$number (int inletnumber, const struct PDdata * message) {
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static struct PDdata DSPtrace2={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache = NULL;
    float * writer;
    char * write_is_dsp;
    int i;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr1;
    if (NULL == DSPtrace2.dsp_stream) DSPtrace2.dsp_stream = object${number}_upstr2;
    if (NULL == cache) cache = object${number}_cache;

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* we always have a dsp rate input on left */
            if (message->sample_count != DSPtrace1.sample_count ||
                message->sample_count != DSPtrace2.sample_count) {
                memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
                memset(DSPtrace2.dsp_stream,0,block_size * sizeof(float));
                DSPtrace1.sample_count = DSPtrace2.sample_count = message->sample_count;
                $reverseDSPcalls
                for (i=0;i<block_size;i++) {
                    cache[i]=DSPtrace1.dsp_stream[i]*DSPtrace2.dsp_stream[i];
                }
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 1: /* inlet 2 = adjust scalar */
            error("(#$number) this *~ takes only audio in right inlet (got \%d)",message->type);
            break;
        case 0: /* inlet 1 = invalid */
            error("(#$number) *~ takes only audio in left inlet (got \%d)",message->type);
            break;
        default:
            error("(#$number) *~ only has two inlets (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for *~ $number */
  object${number}_upstr1 = malloc(block_size*sizeof(float));
  object${number}_upstr2 = malloc(block_size*sizeof(float));
  object${number}_cache = malloc(block_size*sizeof(float));

  if (object${number}_upstr1 == NULL || object${number}_upstr2 == NULL || object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (*~)!");
    exit(1);
  }
];
	 }
       }
	     },
     'random' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::random';
	 my ($number, %myself, @error) = @_;
	 my @arguments = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 # random: two inlets, one outlet
	 #  Inlet 0: Takes bang or [seed n<
	 #  Inlet 1: Takes int (max value+1), strictly non-negative
	 #  Creation argument: Inlet 1.
	 #  Outlet: Random integers from 0 .. max value

	 print STDERR "PuDaC uses a different PRNG than PD; be careful of seeds\n"
	   unless ($PuDaC_PRNG_warning++);

	 if ($arguments[0] <= 0) {
	   print STDERR "random probably doesn't do what you want, given a negative creation argument (sets to 0)\n"
	     if (length $arguments[0]);
	   $arguments[0]=0;
	 }

	 print qq[
/* object $number is a RANDOM object */
void object$number (int inletnumber, const struct PDdata * message) {
    unsigned static int randseed=0xDEADBEEF, maxval = $arguments[0];
    struct PDdata outvalue;

    switch (inletnumber) {
	case 0: /* bang/msg inlet */
            switch (message->type) {
                case PDC_TYPE_BANG:
                    outvalue.linked = NULL;
                    outvalue.type = PDC_TYPE_FLOAT;
                    if (maxval)
                        outvalue.val_float = rand_r(&randseed) % maxval;
                    else
                        outvalue.val_float = 0;
                    $attachedobjectcalls
                    break;
                case PDC_TYPE_SYMBOL:
                    if (strcasecmp("seed",message->val_symbol)==0) {
                        if (message->linked && message->linked->type == PDC_TYPE_FLOAT) {
                            randseed = message->linked->val_float;
                        } else
                            error("(#$number) random doesn't understand message 'seed [nonfloat]'");
                    } else
                        error("(#$number) random doesn't understand message starting with \%s",message->val_symbol);
                    return;
                default:
                    error("(#$number) unknown type \%d", message->type);
                    return;
                }
            break;
        case 1: /* maxval inlet */
            if (message->type != PDC_TYPE_FLOAT) {
	        error("(#$number) random takes only takes float in right inlet");
	        return;
            }
            if (message->val_float < 2)
                maxval = 0;
            else
	        maxval = message->val_float;
            break;
	default: /* not an inlet */
	    error("(#$number) random only has one inlet");
            return;
    }
    return;
}
];
}
	  },
     'mtof' => { 'CODE' => sub {
        local *__ANON__ = 'PuDaC::Objects::mtof';
	my ($number, %myself, @error) = @_;
	my @arguments = @{$myself{ARGS}};
	my $attachedobjectcalls = $myself{AOCS};

	print qq[
/* object $number is an MTOF object */
void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata outvalue;

    outvalue.linked = NULL;
    if (message->type != PDC_TYPE_FLOAT || inletnumber != 0)
        return;

    outvalue.type = PDC_TYPE_FLOAT;
    outvalue.val_float = 440*exp2f((message->val_float-69)/12); /* a440 is midi 69 */

    $attachedobjectcalls
}
];
}
	  },
     'tabread' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::tabread';
	 my ($number, %myself, @error) = @_;
	 my ($name) = @{$myself{ARGS}};
	 my $attachedobjectcalls = $myself{AOCS};

	 print qq[
/* object $number is a TABREAD object */

void object$number (int inletnumber, const struct PDdata * message) {
    struct PDdata totable={PDC_TYPE_SYMBOL,0,"\\\\getguts",0,0,NULL,NULL};
    struct PDdata outvalue;
    static int tablesize=0;
    static float * table = NULL;
    float * writer;
    int i;

    if (NULL == table) {
      sendbyname("$name",&totable);
      table = totable.dsp_stream;
      tablesize = totable.sample_count; /* This is bad. We can't dynamically resize things */
    }

    if (inletnumber != 0) return;

    switch(message->type) {
        case PDC_TYPE_FLOAT:
            outvalue.linked=NULL; outvalue.type=PDC_TYPE_FLOAT;
            if (message->val_float < 0)
                 outvalue.val_float = table[0];
            else if (message->val_float >= tablesize)
                 outvalue.val_float = table[tablesize-1];
            else outvalue.val_float = table[(int)message->val_float];

            $attachedobjectcalls
            break;
        case PDC_TYPE_SYMBOL:
            error("(#$number) Tabread does not yet support 'set newname'"\n);
            break;
        default:
            error("(#$number) tabread takes only floats in inlet (got \%d)",message->type);
    }
}
];
       }
	       },
     'lop~' => { 'CODE' => sub {
	 local *__ANON__ = 'PuDaC::Objects::lop~~';
	 my ($number, %myself, @error) = @_;
	 my ($name) = @{$myself{ARGS}};
	 my $reverseDSPcalls = $myself{DSPCS};

	 # Create a 1st order DT filter, a pole somewhere between z=0 and z=1.
	 # i.e. H=(1-k)/(1-kq); k=1 -> pole at z=1; k=0 -> pole at 0
	 # zero always at 0
	 # Y/X=(1-k)/(1-kq); Y-Ykq=X(1-k); y[n]=(1-k)*x[n]+y[n-1]*k
	 # PD uses k= freq * 2pi / sample_rate
	 # and calculates the block as
	 #  last = *out++ = coef * *in++ + feedback * last

	 print qq[
/* object $number is a LOP~ object */
float *object${number}_upstr, *object${number}_cache;

void object$number (int inletnumber, const struct PDdata * message) {
    static struct PDdata DSPtrace1={PDC_DSP_REQUEST,0,NULL,-1,0,NULL,NULL};
    static float * cache = NULL;
    float last;
    float k;
    float * writer;
    static unsigned int pos=0;
    int i;
    char * write_is_dsp;

    if (NULL == DSPtrace1.dsp_stream) DSPtrace1.dsp_stream = object${number}_upstr;
    if (NULL == cache) cache = object${number}_cache;

    writer = message->dsp_stream; /* break const-ness of message */

    switch (inletnumber) {
        case -1: /* outlet 1 */
            if (message->type != PDC_DSP_REQUEST)
                return;
            write_is_dsp = &message->is_dsp;
            *write_is_dsp=1;
            /* find out if we have a dsp rate input */
            if (message->sample_count != DSPtrace1.sample_count) {
                memset(DSPtrace1.dsp_stream,0,block_size * sizeof(float));
                DSPtrace1.sample_count = message->sample_count;
                $reverseDSPcalls
                for (i=0;i<block_size;i++) {
                    last=cache[i]=(1-k)*DSPtrace1.dsp_stream[i]+k*last;
                }
            }
            for (i=0;i<block_size;i++) {
                writer[i]+=cache[i];
            }
            break;
        case 1: /* inlet 2 = corner frequency */
            if (message->type == PDC_TYPE_FLOAT) {
                if (message->val_float < 0) pos = 0;
                else if (message->val_float >= tablesize) pos=tablesize-1;
                else k = message->val_float * 2*M_PI / sample_rate;
            } else
                error("(#$number) lop~ takes only floats in freq inlet (got \%d)",message->type);

            break;
        default:
            error("(#$number) lop~ only has one non-dsp inlet (tried \%d)",inletnumber);
            return;
    }
}
];#]]]]

	 $globalinit .= qq[
  /* allocate "private" memory for lop~ $number */
  object${number}_upstr = malloc(block_size*sizeof(float));
  object${number}_cache = malloc(block_size*sizeof(float));

  if (object${number}_upstr == NULL || object${number}_cache == NULL) {
    error("(#$number) Unable to allocate memory for obj $number (lop~)!");
    exit(1);
  }
];
       }
	       },
    );

1;

# objects done:
#  == + - * / > < int/i float/f floatatom send/s receive/r max min sqrt abs mtof
#  loadbang table(half) print metro delay timer until bang/b \msg random
#  spigot clip trigger unpack pack select route toggle/tgl bng
#  osc~ dac~ tabread~ line~
#  netreceive

# still need

# route and select have two forms each (annoyingly):
#  one that deals with symbols, and one that deals with floats.
#  as far as i can tell, there's no good reason for this.
